/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * include/commands/diagnostic_commands.h
 *
 * Exports for the diagnostic commands.
 *
 *-------------------------------------------------------------------------
 */

#ifndef DIAGNOSTIC_COMMANDS_H
#define DIAGNOSTIC_COMMANDS_H

extern Datum current_op(pgbson *commandSpec);

#endif
